%{
   open Syntax
   open Support.Error
     
   let actual = ref "No further parse error information available"
   let parse_error s = 
     raise (Failure (!actual))
     (* print_string ("\n\n Details: "^(!actual)^".\n\n") *)

   let appexpcount = ref 0
   let make_appexpid () =
     appexpcount := !appexpcount+1;
     !appexpcount
%}

/* 

   Authors:  Steffen Jost  <jost@informatik.uni-muenchen.de>
	     using a template file from
	     David Aspinall	<da@dcs.ed.ac.uk> and
	     Martin Hofmann	<mhofmann@informatik.uni-muenchen.de> 
             Brian Campbell

   This is a modified version of the normal parser which is closer to
   the syntax in my (Brian's) thesis.

*/

/* Ordered by appearance as in lexer.mll */
%token <Support.Error.info> EOF
%token <int Support.Error.withinfo> DIASIZE
%token <string Support.Error.withinfo> 	IDENTIFIER
%token <string Support.Error.withinfo> 	CONSTRUCTOR
/* KEYWORDS ARE DEFINED BELOW */
%token <Support.Error.info> 		UNITVAL
%token <int    Support.Error.withinfo> 	INTVAL
%token <float  Support.Error.withinfo> 	FLOATVAL
%token <char   Support.Error.withinfo> 	CHARVAL
%token <string Support.Error.withinfo> 	STRINGVAL
/* Den Unterschied zwischen QUOTEVAL und STRINGVAL in Camelot habe ich nicht verstanden.
   Deshalb ignoriere ich QUOTEVAL. 
   Das STRINGVAL als ein beliebiges unidentifiziertes KEYWORD definiert ist,
   finde ich auch ziemlich komisch, das mache ich anders, also so wie QUOTEVAL,
   so das ich letztendlich also STRINGVAL und nicht QUOTEVAL ignoriere! 
*/
%token <Support.Error.info> PRIME
%token <Support.Error.info> EQUALS
%token <Support.Error.info> LESS
%token <Support.Error.info> GREATER
%token <Support.Error.info> DIAMOND
%token <Support.Error.info> LTEQ
%token <Support.Error.info> GTEQ
%token <Support.Error.info> ARROW
%token <Support.Error.info> WARROW	
%token <Support.Error.info> LPAREN
%token <Support.Error.info> RPAREN
%token <Support.Error.info> EMPTYPAREN
%token <Support.Error.info> LBRACE
%token <Support.Error.info> RBRACE
%token <Support.Error.info> LBRAK
%token <Support.Error.info> RBRAK
%token <Support.Error.info> COMMA
%token <Support.Error.info> COLON 
%token <Support.Error.info> SEMICOLON
/* AND is also a keyword */
%token <Support.Error.info> BAR
%token <Support.Error.info> AT
%token <Support.Error.info> BANG
%token <Support.Error.info> DOLLAR
%token <Support.Error.info> SHARP
%token <Support.Error.info> POUND
%token <Support.Error.info> PARAGRAPH
/* %token <Support.Error.info> USCORE   This token was removed. */
%token <Support.Error.info> CARET
%token <Support.Error.info> PLUS 
%token <Support.Error.info> MINUS
%token <Support.Error.info> TIMES
%token <Support.Error.info> DIVIDE
%token <Support.Error.info> FPLUS 
%token <Support.Error.info> FMINUS
%token <Support.Error.info> FTIMES
%token <Support.Error.info> FDIVIDE
%token UMINUS
/* KEYWORDS */
%token <Support.Error.info> UNIT
%token <Support.Error.info> BOOL
%token <Support.Error.info> INTEGER
%token <Support.Error.info> FLOAT
%token <Support.Error.info> ARRAY
%token <Support.Error.info> CHAR
%token <Support.Error.info> STRING

%token <Support.Error.info> TYPE
%token <Support.Error.info> VAL
%token <Support.Error.info> ANNVAL
%token <Support.Error.info> FULLVAL
%token <Support.Error.info> OF

%token <Support.Error.info> TRUEVAL
%token <Support.Error.info> FALSEVAL

%token <Support.Error.info> BOOL
%token <Support.Error.info> NOT
%token <Support.Error.info> LAND
%token <Support.Error.info> LOR
%token <Support.Error.info> ANDALSO
%token <Support.Error.info> ORELSE
%token <Support.Error.info> MOD

%token <Support.Error.info> BEGIN
%token <Support.Error.info> END
%token <Support.Error.info> IF
%token <Support.Error.info> THEN
%token <Support.Error.info> ELSE
%token <Support.Error.info> MATCH
%token <Support.Error.info> MATCHPR
%token <Support.Error.info> WITH
%token <Support.Error.info> LET
%token <Support.Error.info> AND
%token <Support.Error.info> REC
%token <Support.Error.info> IN
%token <Support.Error.info> WHERE
%token <Support.Error.info> FUN

/* Das sind eigentlich keine Keywords, sondern Built-in functions die ich
   erst bei der Prfung der Identifier abfange... 
 %token <Support.Error.info> INT_OF_FLOAT
 %token <Support.Error.info> FLOAT_OF_INT
 %token <Support.Error.info> CHAR_OF_INT
 %token <Support.Error.info> INT_OF_CHAR
 %token <Support.Error.info> FLOAT_OF_STRING
 %token <Support.Error.info> STRING_OF_FLOAT
 %token <Support.Error.info> INT_OF_STRING
 %token <Support.Error.info> STRING_OF_INT
 %token <Support.Error.info> PRINT_OF_INT
 %token <Support.Error.info> PRINT_OF_FLOAT
 %token <Support.Error.info> PRINT_OF_CHAR
 %token <Support.Error.info> PRINT_OF_STRING
 %token <Support.Error.info> ARRAY_HEAD
*/


/* Operatorsymbole in aufsteigender Prioritt */
%nonassoc LET AND REC IN
%nonassoc MATCH MATCHPR WITH WARROW
%nonassoc IF THEN ELSE
%nonassoc COLON
%nonassoc ARRAY
%left WHERE
%right SEMICOLON COMMA
%right ARROW
%left ANDALSO ORELSE
%left LAND LOR  
%right NOT
%nonassoc EQUALS LESS GREATER LTEQ GTEQ
%left CARET /* String append */
%left PLUS MINUS FPLUS FMINUS
%left TIMES DIVIDE FTIMES FDIVIDE 
%nonassoc MOD
%nonassoc UMINUS UFMINUS


/* Starting symbols (must have type) */
%start pprogram
%type <Syntax.program> pprogram


%%

typname: /* Names of user-types may be upper or lower case */ 
	    IDENTIFIER   { $1 }
	|   CONSTRUCTOR  { $1 }
;

pprogram: 
	  ptypdeclist pvaldeclist pfundeflist EOF { Program($4,$1,$2,$3) }
;

ptypdeclist:
					{ [] }
	| ptypdec ptypdeclist		{ $1 :: $2 }
;

ptypdec:
	  TYPE typname EQUALS pconstructordeclist psemicolonopt
		{ TypDec($1,[],$2.v,$4) }
;

pconstructordeclist:	
	  pconstructordec				{ $1 :: [] }
	| pconstructordec BAR pconstructordeclist	{ $1 :: $3 }
;

pconstructordec:		
	   CONSTRUCTOR pconstrsize 				{ (if $2 = Syntax.Unspecified then Support.Error.warningAt $1.i "Unspecified Constructor size, using default size.");
                                                                  TypCon($1.i,$1.v,$2,[]) }
	|  BANG CONSTRUCTOR 	                                { TypCon($2.i,$2.v,Syntax.Specified(0),[]) }
	|  CONSTRUCTOR pconstrsize OF psimpletypelist	        { (if $2 = Syntax.Unspecified then Support.Error.warningAt $1.i "Unspecified Constructor size, using default size.");
								  TypCon($1.i,$1.v,$2,$4) }
;


pconstrsize:
	                        { Syntax.Unspecified }
	| DIASIZE               { Syntax.Specified($1.v) }
;

pvaldeclist:
				{ [] }
	| pvaldec pvaldeclist	{ $1 :: $2 }
;

pvaldec:  
	  VAL typname pconstrsize COLON pfunctiontype psemicolonopt 	
	    { ValDec($1,$2.v,$3,$5) }
	| ANNVAL typname pconstrsize COLON pdianumopt COMMA prfunctiontype COMMA pdianumopt psemicolonopt
	    { AnnValDec($1,$2.v,$3,$5,$7,$9) }
	| FULLVAL typname pconstrsize COLON pidordia COMMA prfunctiontype COMMA pidordia LBRACE constraints RBRACE psemicolonopt
            { FullValDec($1,$2.v,$3,$5,$7,$9,$11) }
;

psemicolonopt: /* There might be a semicolon, but it is not necessary */
	            { }
	| SEMICOLON { }
;

pdianum: /* Deprecated. Move into pdianumopt! */
	  LESS FLOATVAL GREATER { $2.v }
	| LESS   INTVAL GREATER { float_of_int($2.v) }
;

pdianumopt:
          pdianum             { Some $1 }
	| TIMES               { None    }
	| LESS TIMES GREATER  { None    } /* For convenience */
;

pidordia:
	  IDENTIFIER            { $1.v }
        | pdianumopt		{ match $1 with None -> "None"
                                    | Some v -> string_of_float v }

psimpletype: psimpletype_ { actual:= "Expecting a simple type here.";$1 }
psimpletype_:
	  pbasetype			{ $1 }
        | psimpletype TIMES psimpletype { {i=$2; v=PairTyp($1,$3)} }
	| psimpletype ARRAY		{ {i=$2; v=ArrayTyp($1)} }      /* Arrays kennen wir eigentlich nicht! */
	| typname 	                { {i=$1.i; v=ConTyp([],$1.v)} } /* Generische Konstruktor typen */
;

psimpletypelist:
          psimpletype                   { [$1] }
        | psimpletype COMMA psimpletypelist { $1::$3 }

pfunctiontype:
          psimpletypelist ARROW psimpletype { Syntax.curry($1@[$3]) }

pbasetype: pbasetype_ { actual:= "Expecting a base type here.";$1 }
pbasetype_:
	  UNIT				{ {i=$1; v=UnitTyp} }
	| DIAMOND			{ {i=$1; v=DiamondTyp} }
	| BOOL				{ {i=$1; v=BoolTyp} }
	| INTEGER			{ {i=$1; v=IntTyp} }
	| FLOAT				{ {i=$1; v=FloatTyp} }
	| CHAR				{ {i=$1; v=CharTyp} }
	| STRING			{ {i=$1; v=StringTyp} }
;

prtype: /* Syntax.rich_type - does not contain any 'info'-elements 
           The idea is, that the user already states the wanted rich_type.
	   However, not all necessary informations are already computed,
	   hence the rich_type is not really valid and must be completed
           by Constraint.complete_annval_rt. This is a hack, yes. :-(
	*/
	  psubrbasetype			{ $1 }
	| typname LBRAK pconrtlist RBRAK	       
                                        { RConTyp([],$1.v,$3) } 
    /*	| prtype ARRAY	      	        { RArrayTyp($1) }         Arrays are not implemented yet! */
;


prtypelist:
          prtype                        { [$1] }
        | prtype COMMA prtypelist       { $1::$3 }

prfunctiontype:
          prtypelist ARROW prtype       { Syntax.rich_curry($1@[$3]) }

prrectype: /* A limited prtype, where constructor types are not allowed to have parameters in front. 
              Saves some parenthesis in the source code. */
   typname LBRAK pconrtlist RBRAK       { RConTyp([],$1.v,$3) } 
;


pconrtlist: /* Syntax.rt_contab */
	  pconrtarg pconrtlistend
	    {
 	     let (cnstr,rcinf) = $1 in
	     ConTab.add cnstr rcinf $2 
	   }
;
pconrtlistend: /* Syntax.rt_contab */
	                 { ConTab.empty }
	| BAR pconrtlist { $2 }
;

pconrtarg: /* (constructor * Syntax.rich_coninfo) */
          CONSTRUCTOR LPAREN pconrtarglist RPAREN
            { let (args, d) = $3 in 
	      ($1.v,
	       {
		(* Must be converted back into a proper constant -> lp_solve does not accept linear arithmetic expressions as factors *)
		rcvar = d;
                rgiveback = "";
		rorder = 0;                (* UNKNOWN!   0 denotes unknown order *)
		rarg_types = args;         (* May contain incomplete subdata *)
		rsize = Unspecified        (* UNKNOWN! *)  
	      }
	      )
	    }
;

pconrtarglist: /* (rich_typ list * float ) */
          pidordia                           { ([], $1) }
	| psubrtypeself COMMA pconrtarglist  { let (args, d) = $3 in (($1::args), d) }
;


psubrbasetype: psubrbasetype_ { actual:= "Expecting a base type here. Rich types must be enclosed in parenthesis.";$1 }
psubrbasetype_:
	  UNIT				{ RUnitTyp }
	| DIAMOND			{ RDiamantTyp }
	| BOOL				{ RBoolTyp }
	| INTEGER			{ RIntTyp }
	| FLOAT				{ RFloatTyp }
	| CHAR				{ RCharTyp }
	| STRING			{ RStringTyp }
	| LPAREN prtype RPAREN		{ $2 }
;


psubrtypeself: 
          SHARP                         { rSelfTypUnknown }
	| prrectype                     { $1 }
	| psubrbasetype                 { $1 }
;


/* END of the ANNVAL-RICHTYPE-HACK */

constraints:
	                                   { [] }
	| aconstraint SEMICOLON constraints { $1 :: $3 }
;

plusminopt:
	        {  1.0 }
	| PLUS  {  1.0 }
	| MINUS { -1.0 }

plusmin:
	  PLUS  {  1.0 }
	| MINUS { -1.0 }

lincomb:
	  plusminopt INTVAL lincombrem { ($1*.(float_of_int $2.v), "K1")::$3 }
	| plusminopt FLOATVAL lincombrem { ($1*.$2.v, "K1")::$3 }
        | plusminopt INTVAL IDENTIFIER lincombrem { ($1*.(float_of_int $2.v), $3.v)::$4 }
        | plusminopt FLOATVAL IDENTIFIER lincombrem { ($1*.$2.v, $3.v)::$4 }
        | plusminopt IDENTIFIER lincombrem { ($1, $2.v)::$3 }

lincombrem:
	  { [] }
	| plusmin INTVAL lincombrem { ($1*.(float_of_int $2.v), "K1")::$3 }
	| plusmin FLOATVAL lincombrem { ($1*.$2.v, "K1")::$3 }
        | plusmin INTVAL IDENTIFIER lincombrem { ($1*.(float_of_int $2.v), $3.v)::$4 }
        | plusmin FLOATVAL IDENTIFIER lincombrem { ($1*.$2.v, $3.v)::$4 }
        | plusmin IDENTIFIER lincombrem { ($1, $2.v)::$3 }

aconstraint:
	  lincomb EQUALS lincomb  { UserEq($1,$3) }
        | lincomb LTEQ   lincomb  { UserGeq($3,$1) }
        | lincomb GTEQ   lincomb  { UserGeq($1,$3) }

pfundeflist:
	  				{ [] }
	| pfundef pfundeflist 		{ $1::$2 }	
;

pfundef:
	  LET precopt IDENTIFIER pvarlist EQUALS pexpression pandfundef	{ FunctionDef($1,$3.v,$4,$6) :: $7 }
;

pandfundef:
									{ [] }
	| AND         IDENTIFIER pvarlist EQUALS pexpression pandfundef	{ FunctionDef($1,$2.v,$3,$5) :: $6 }
;

precopt: /* Es darf ein REC kommen, muss aber nicht */
		{ }
	| REC	{ }
;

pexpression:
          plainexpression                       { $1 }
        | LPAREN plainexpression RPAREN         { $2 }
	| pvalue                                { {i=$1.i; v=ValueExp($1)} }
;

/* 
   Die obige Unterteilung ist notwendig, da bei aufeinanderfolgenden Matches die jeweiligen 
   Matchrules nicht korrekt zugeordnet werden knnen...
*/


pbaropt: /* Hier darf ein BAR kommen, muss aber nicht */
		{ }
	| BAR	{ }
;

anyarrow:
     WARROW     { $1 }  /* For compatibility with OCAML... */
   |  ARROW     { $1 }  /* ...and for convenience.         */

pmatchrulelist:
	  pmatchrule				{ [$1] }
	| pmatchrule BAR pmatchrulelist		{ $1::$3 }
;

pmatchrule: pmatchrule_ { actual:= "Expecting a constructor for a pattern match rule here.";$1 }
pmatchrule_:
      CONSTRUCTOR pvarargs PRIME anyarrow pexpression   { Matchrule($4,$1.v,$2,true ,$5,Syntax.Void) } 
   |  CONSTRUCTOR pvarargs anyarrow pexpression   { Matchrule($3,$1.v,$2,false,$4,Syntax.New) } 
;


plainexpression:
	  MATCH IDENTIFIER WITH pbaropt pmatchrulelist  	{ {i=$1; v=MatchExp($2.v,$5)}  }
	| MATCH IDENTIFIER WITH LPAREN IDENTIFIER COMMA IDENTIFIER RPAREN anyarrow pexpression
                                                { {i=$1; v=MatchPExp($2.v,$5.v,$7.v,$10)}  }
        | IF pvalue THEN pexpression ELSE pexpression
                                                { {i=$1; v=IfExp($2,$4,$6)} }
	| LET IDENTIFIER EQUALS pexpression IN pexpression
					        { if $2.v = "_" 
						  then {i=$1; v=SeqExp($4,$6)} 
						  else {i=$1; v=LetExp($2.v,$4,$6,[])} 
                                                }   /* Deprecated, as lexer does not identify USCORE anymore: | LET USCORE EQUALS pexpression IN psubexpression { {i=$1; v=SeqExp($4,$6)} } */
	| LET IDENTIFIER LPAREN pvarlist RPAREN EQUALS pexpression IN pexpression 
					        { if $2.v = "_" 
						  then {i=$1; v=SeqExp($7,$9)} 
						  else {i=$1; v=LetExp($2.v,$7,$9,$4)} 
						}
	| pexpression SEMICOLON pexpression	
	                                        { {i=$2; v=SeqExp($1,$3)} }
	| IDENTIFIER  funarglist                { {i=$1.i; v=AppExp($1.v,$2,make_appexpid ())} }        /* Function call */
	| CONSTRUCTOR pvalueargs  	        { {i=$1.i; v=ConstrExp($1.v,$2,Syntax.New)} }     /* Constructor application */
;

pvalue:
	  patomicvalue                       { $1 }
        | LPAREN pvalue COMMA pvalue RPAREN  { {i=$1; v=PairVal($2,$4)} }
	| LPAREN pvalue RPAREN               { $2 }
	| MINUS	 %prec UMINUS  pvalue        { {i=$1; v=UnaryOpVal({i=$1; v=UMinusOp },$2) } }
	| FMINUS %prec UFMINUS pvalue        { {i=$1; v=UnaryOpVal({i=$1; v=UFminusOp},$2) } }
	| NOT  		       pvalue        { {i=$1; v=UnaryOpVal({i=$1; v=NotOp},$2) } }
	| pvalue TIMES   pvalue              { {i=$2; v=BinaryOpVal({i=$2; v=TimesOp  },$1,$3)} } 
	| pvalue DIVIDE  pvalue              { {i=$2; v=BinaryOpVal({i=$2; v=DivOp    },$1,$3)} } 
	| pvalue PLUS    pvalue              { {i=$2; v=BinaryOpVal({i=$2; v=PlusOp   },$1,$3)} } 
	| pvalue MINUS   pvalue              { {i=$2; v=BinaryOpVal({i=$2; v=MinusOp  },$1,$3)} } 
	| pvalue FTIMES  pvalue              { {i=$2; v=BinaryOpVal({i=$2; v=FtimesOp },$1,$3)} } 
	| pvalue FDIVIDE pvalue              { {i=$2; v=BinaryOpVal({i=$2; v=FdivOp   },$1,$3)} } 
	| pvalue FPLUS   pvalue              { {i=$2; v=BinaryOpVal({i=$2; v=FplusOp  },$1,$3)} } 
	| pvalue FMINUS  pvalue              { {i=$2; v=BinaryOpVal({i=$2; v=FminusOp },$1,$3)} } 
	| pvalue LESS    pvalue              { {i=$2; v=BinaryOpVal({i=$2; v=LessOp   },$1,$3)} } 
	| pvalue LTEQ    pvalue              { {i=$2; v=BinaryOpVal({i=$2; v=LteqOp   },$1,$3)} } 
	| pvalue GREATER pvalue              { {i=$2; v=BinaryOpVal({i=$2; v=GreaterOp},$1,$3)} } 
	| pvalue GTEQ    pvalue              { {i=$2; v=BinaryOpVal({i=$2; v=GteqOp   },$1,$3)} } 
	| pvalue EQUALS  pvalue              { {i=$2; v=BinaryOpVal({i=$2; v=EqualOp  },$1,$3)} } 
	| pvalue CARET   pvalue              { {i=$2; v=BinaryOpVal({i=$2; v=AppendOp },$1,$3)} } 
	| pvalue LAND    pvalue              { {i=$2; v=BinaryOpVal({i=$2; v=AndOp    },$1,$3)} } 
	| pvalue LOR     pvalue              { {i=$2; v=BinaryOpVal({i=$2; v=OrOp     },$1,$3)} } 
	| pvalue ANDALSO pvalue              { {i=$2; v=BinaryOpVal({i=$2; v=AndalsoOp},$1,$3)} } 
	| pvalue ORELSE  pvalue              { {i=$2; v=BinaryOpVal({i=$2; v=OrelseOp },$1,$3)} } 
	| pvalue MOD     pvalue              { {i=$2; v=BinaryOpVal({i=$2; v=ModOp },$1,$3)} } 
;


patomicvalue:
	  IDENTIFIER			{ {i=$1.i; v=VarVal($1.v)} } /* Could also be a 0-arity function call, but should not be! */
	| INTVAL			{ {i=$1.i; v=IntVal($1.v)} }
	| FLOATVAL			{ {i=$1.i; v=FloatVal($1.v)} }
	| CHARVAL			{ {i=$1.i; v=CharVal($1.v)} }
	| STRINGVAL			{ {i=$1.i; v=StringVal($1.v)} }
	| TRUEVAL			{ {i=$1; v=BoolVal(true)} }
	| FALSEVAL			{ {i=$1; v=BoolVal(false)} } 
	| UNITVAL			{ {i=$1; v=UnitVal} }        
;


/*
 For compatibility with CAMELOT, we must distinguish to kinds of argument lists:
 	list - lists like " a b c d ", " a ", " " 
   	args - lists like "(a,b,c,d)", "(a)", "()" or " "
 note that an arg-list consiting of a single unit argmuent must be written as "(())" 
*/

pvaluelist:
       pvalue                   { [$1] }
   |   pvalue COMMA pvaluelist	{ $1::$3 }
;

funarglist:
       patomicvalue             { [$1] }
     | LPAREN pvaluelist RPAREN { $2 }                    
;

pvalueargs:
	                                       	 	{ [] }
	| EMPTYPAREN                           	 	{ [] }
	| UNITVAL                              	 	{ [] } /* Special Case: if there is a single argument of type unit, it must be enclosed in parenthesis in an argument-list */	
	| LPAREN RPAREN                       	 	{ [] }
	| LPAREN pvalue pvalueargsaux RPAREN	{ $2 :: $3 }
;

pvalueargsaux:
	      					{ [] } 	
	| COMMA pvalue pvalueargsaux  	{ $2 :: $3 }
;

commaidlist:
          IDENTIFIER                    { [$1.v] }
        | IDENTIFIER COMMA commaidlist  { $1.v :: $3 }

pvarlist:
	  				{ [] }
        | IDENTIFIER                    { [$1.v] }
        | LPAREN commaidlist RPAREN     { $2 }
;

pvarargs: 
	                                        { [] }
	| EMPTYPAREN                            { [] }
	| UNITVAL                               { [] } /* Special Case: if there is a single argument of type unit, it must be enclosed in parenthesis in an argument-list */	
	| LPAREN RPAREN                         { [] }
	| LPAREN IDENTIFIER pvarargsaux RPAREN 	{ $2.v :: $3 }
;

pvarargsaux:
	      				{ [] } 	
	| COMMA IDENTIFIER pvarargsaux  { $2.v :: $3 }
;



